/* gpio.c - Routines to access RPi GPIO for joystick emulation.

Strongly based upon Gertboard test software:
   Copyright (C) Gert Jan van Loo & Myra VanInwegen 2012
   No rights reserved
   You may treat this program as if it was in the public domain

*/

#include "gpio.h"

#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO   */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

int  mem_fd;
char *gpio_mem_orig, *gpio_mem, *gpio_map;

// I/O access
volatile unsigned *gpio;

//
//  GPIO
//

// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET0   *(gpio+7)  // Set GPIO high bits 0-31
#define GPIO_SET1   *(gpio+8)  // Set GPIO high bits 32-53

#define GPIO_CLR0   *(gpio+10) // Set GPIO low bits 0-31
#define GPIO_CLR1   *(gpio+11) // Set GPIO low bits 32-53

#define GPIO_IN0   *(gpio+13)  // Reads GPIO input bits 0-31

#define GPIO_PULL   *(gpio+37) // Pull up/pull down
#define GPIO_PULLCLK0 *(gpio+38) // Pull up/pull down clock

//
// This is a software loop to wait
// a short while.
//
void short_wait()
{ int w;
  for (w=0; w<100; w++)
  { w++;
    w--;
  }
} // short_wait

//
// Set up memory regions to access the peripherals.
// This is a bit of 'magic' which you should not touch.
// It it also the part of the code which makes that
// you have to use 'sudo' to run this program.
//
int gpio_init (void)
{
  unsigned long extra;

   /* open /dev/mem */
  if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0)	 return	  GPIO_ERR_MEM;

   /*
    * mmap GPIO
    */
   if ((gpio_mem_orig = malloc(BLOCK_SIZE + (PAGE_SIZE-1))) == NULL) return	  GPIO_ERR_ALLOC;
   extra = (unsigned long)gpio_mem_orig % PAGE_SIZE;
   if (extra)
     gpio_mem = gpio_mem_orig + PAGE_SIZE - extra;
   else
     gpio_mem = gpio_mem_orig;

   gpio_map = (unsigned char *)mmap(
      (caddr_t)gpio_mem,
      BLOCK_SIZE,
      PROT_READ|PROT_WRITE,
      MAP_SHARED|MAP_FIXED,
      mem_fd,
      GPIO_BASE
   );

   if ((long)gpio_map < 0) return	GPIO_ERR_MAP;
   gpio = (volatile unsigned *)gpio_map;

   return	GPIO_OK;
}

void gpio_term (void)
	{
	munmap(gpio_map,BLOCK_SIZE);
	// free memory
	free(gpio_mem_orig);
	}

void gpio_input (int iMask)
	{
	int	iPin;
	for ( iPin = 0; iPin < 32; ++iPin )
		{
		if ( iMask & 1 )
			{
			INP_GPIO (iPin);
			}
		iMask  /= 2;
		}
	}

void gpio_pullup (int iMask)
	{
	// enable pull-up.
	GPIO_PULL = 2;
	short_wait();
	GPIO_PULLCLK0 = iMask;
	short_wait();
	GPIO_PULL = 0;
	GPIO_PULLCLK0 = 0;
	}

void gpio_pullnone (int iMask)
	{
	// enable pull-up.
	GPIO_PULL = 0;
	short_wait();
	GPIO_PULLCLK0 = iMask;
	short_wait();
	GPIO_PULL = 0;
	GPIO_PULLCLK0 = 0;
	}

int gpio_get (int iMask)
	{
	return	iMask & GPIO_IN0;
	}
